--[[
    Database of all battle pets and mounts
]]--
--[[
Copyright (c) 2008-2014, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

-- $Revision: 268 $
-- $Date: 2015-01-19 18:00:12 +0000 (Mon, 19 Jan 2015) $

local MAJOR_VERSION = "LibLordFarlander-PetAndMountDatabase-2.0"
local MINOR_VERSION = tonumber( ( "$Revision: 268 $" ):match("(%d+)" ) ) + 90000

local LibPetAndMountDatabase, oldminor = LibStub:NewLibrary( MAJOR_VERSION, MINOR_VERSION )
if( not LibPetAndMountDatabase ) then
    error( MAJOR_VERSION .. " could not be created!" )
end--if

if( not LibPetAndMountDatabase.FoundSets ) then
    LibPetAndMountDatabase.FoundSets = {}
end--if
if( not LibPetAndMountDatabase.SubSets ) then
    LibPetAndMountDatabase.SubSets = {}
end--if

local LibPeriodicTable = LibStub( "LibPeriodicTable-3.1", true )
if( not LibPeriodicTable ) then
    error( MAJOR_VERSION .. " requires LibPeriodicTable-3.1" )
end--if

local TR = LibStub( "LibLordFarlander-TableRecycler-1.0", true )
if( not TR ) then
    error( MAJOR_VERSION .. " requires LibLordFarlander-TableRecycler-1.0" )
    return
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true )
if( not ItemList ) then
    error( MAJOR_VERSION .. " requires LibLordFarlander-ItemList-2.0" )
    return
end--if

function LibPetAndMountDatabase.NormalizeSetName( set )
    if( set:find( "LibLordFarlander%-PetAndMountDatabase%-3%." ) ~= 1 ) then
        set = "LibLordFarlander-PetAndMountDatabase-3." .. set
    end--if
    return set
end--LibPetAndMountDatabase.NormalizeSetName( set )

function LibPetAndMountDatabase:HasSet( set, normalize )
    local sets = LibPeriodicTable.sets

    if( ( normalize == true ) or ( normalize == nil ) ) then
        set = self.NormalizeSetName( set )
    end--if
    if( self.FoundSets[set] ) then
        return true
    end--if
    for setName, _ in pairs( sets ) do
        if( setName:find( set ) == 1 ) then
            self.FoundSets[set] = true
            return true
        end--if
    end--if
    return false
end--LibPetAndMountDatabase:HasSet( set, normalize )

function LibPetAndMountDatabase:GetSubsets( set, normalize )
    local sets = LibPeriodicTable.sets

    if( ( normalize == true ) or ( normalize == nil ) ) then
        set = self.NormalizeSetName( set )
    end--if
    if( self.SubSets[set] ) then
        return self.SubSets[set]
    end--if
    set = set .. "."
    for setName, _ in pairs( sets ) do
        local start, ends = setName:find( set )

        if( ( start == 1 ) and ( ends ~= setName:len() ) ) then
            setName = setName:sub( ends + 1 )
            if( setName:find( "." ) ) then
                setName = ("%s%s"):format( set, strsplit( ".", setName ) )
            end--if
            if( self.SubSets[set] == nil ) then
                self.SubSets[set] = {}
            end--if
            self.SubSets[set][setName] = setName
        end--if
    end--if
    return self.SubSets[set]
end--LibPetAndMountDatabase:GetSubsets( set, normalize )

function LibPetAndMountDatabase:IdInSet( item, set, normalize )
    if( item == nil ) then
        return false
    end--if
    if( ( normalize == true ) or ( normalize == nil ) ) then
        set = self.NormalizeSetName( set )
    end--if
    return LibPeriodicTable:ItemInSet( item, set )
end--LibPetAndMountDatabase:IdInSet( item, set, normalize )

function LibPetAndMountDatabase:IterateSet( set, normalize )
    if( ( normalize == true ) or ( normalize == nil ) ) then
        set = self.NormalizeSetName( set, normalize )
    end--if
    return LibPeriodicTable:IterateSet( set )
end--LibPetAndMountDatabase:IterateSet( set, normalize )

function LibPetAndMountDatabase.GetConditions( value )
    local conditions = value:match( "%[(.-)%]" )
    local returnval

    if( conditions ) then
        conditions = TR:GetTableWithInserts( strsplit( ";", conditions ) )
        for _, condition in pairs( conditions ) do
            local name, value = condition:match( "(.-)=(.+)" )

            if( name and value ) then
                if( not returnval ) then
                    returnval = TR:GetTable()
                end--if
                returnval[name] = value
            end--if
        end--for
        TR:ReturnTable( conditions )
    end--if
    return returnval
end--if

function LibPetAndMountDatabase.GetMountAttributes( attributeStrting )
    local mountInfo = TR:GetTable()
    local attributes = TR:GetTableWithInserts( strsplit( ";", attributeStrting ) )

    if( attributes ) then
        for _, attributePair in pairs( attributes ) do
            if( ( type( attributePair ) == "nil" ) or ( type( attributePair ) == "string" ) ) then
                local attribute, value = strsplit( "=", attributePair )

                if( not( value ) or ( value == "true" ) ) then
                    value = true
                elseif( value == "false" ) then
                    value = false
                end--if
                if( type( value ) == "string" ) then
                    if( value:find( "&" ) ) then
                        value = TR:GetTableWithInserts( strsplit( "&", value ) )
                    end--if
                end--if
                if( attribute == "AquaticSpeed" ) then
                    if( type( value ) == "string" ) then
                        value = tonumber( value )
                    else
                        value = nil
                    end--if
                elseif( attribute == "FlyingSpeed" ) then
                    if( type( value ) == "string" ) then
                        value = tonumber( value )
                    else
                        value = nil
                    end--if
                elseif( attribute == "GroundSpeed" ) then
                    if( type( value ) == "string" ) then
                        value = tonumber( value )
                    else
                        value = nil
                    end--if
                elseif( attribute == "Riders" ) then
                    if( type( value ) == "string" ) then
                        value = tonumber( value )
                    else
                        value = nil
                    end--if
                elseif( attribute == "SwimSpeed" ) then
                    if( type( value ) == "string" ) then
                        value = tonumber( value )
                    else
                        value = nil
                    end--if
                -- elseif( attribute == "Zone" ) then
                    -- -- Return the name, not the mapID
                    -- if( type( value ) == "string" ) then
                        -- value = GetMapNameByID( tonumber( value ) )
                    -- elseif( type( value ) == "table" ) then
                        -- local newvalue = TR:GetTable()

                        -- for mapID in pairs( value ) do
                            -- tinsert( newvalue, GetMapNameByID( tonumber( mapID ) ) )
                        -- end--for
                        -- TR:ReturnTable( value )
                        -- value = newvalue;
                    -- end--if
                end--if
                mountInfo[attribute] = value
            end--if
        end--if
        TR:ReturnTable( attributes )
    end--if
    return mountInfo
end--LibPetAndMountDatabase.GetMountAttributes( value )

function LibPetAndMountDatabase:GetPlayersBattleCompanions()
    local returnList = TR:GetTable()
    local numPets, numOwned = C_PetJournal.GetNumPets()
    local BuildBattlePet = ItemList.BuildBattlePet

    for i = 1, numOwned do
        local petID, speciesID, isOwned, customName, level, favorite, isRevoked, creatureName,
              icon, petType, creatureID, sourceText, description, isWildPet, canBattle, isTradeable,
              isUnique, obtainable = C_PetJournal.GetPetInfoByIndex( i )

        if( petID ~= "0x0000000000000000" ) then
            local fullPetID = petID

            petID = tonumber( petID:sub( 3 ), 16 )

            if( isOwned ) then
                local item = BuildBattlePet( petID, fullPetID )

                item.petInfo = TR:GetTable( "modelID", creatureID,
                                            "level", level,
                                            "speciesID", speciesID,
                                            "creatureName", creatureName,
                                            "sourceText", sourceText,
                                            "description", description,
                                            "isWildPet", isWildPet,
                                            "canBattle", canBattle )
                tinsert( returnList, item )
            end--if
        end--if
    end--for
    return returnList;
end--LibPetAndMountDatabase:GetPlayersBattleCompanions()

function LibPetAndMountDatabase:GetPlayersMounts()
    local returnList = TR:GetTable()
    local count = 0
    local BuildMount = ItemList.BuildMount

    for i = 1, C_MountJournal.GetNumMounts() do
        local _, _, _, _, isUsable, _, isFavorite, _, _, hideOnChar, isCollected = C_MountJournal.GetMountInfo( i )

        -- Workaround that isUsable will filter out mounts that can't be used because of the current zone.
        -- We want to figure that out ourselves for caching reasons.
        if( ( mountType == 232 ) or ( mountType == 241 ) or ( mountType == 254 ) ) then
            isUsable = true
        end--if
        if isUsable and not hideOnChar and isCollected then
            local creatureID, _, _, _, mountType = C_MountJournal.GetMountInfoExtra( i )
            local index = #returnList + 1
            local item = BuildMount( i )
            local mountInfo = TR:GetTable( "modelID", creatureID,
                                           "isFavorite", isFavorite )

            item.mountInfo = mountInfo
            tinsert( returnList, item )
            if( mountType ) then
                -- 230 for most ground mounts
                -- 231 for Riding Turtle and Sea Turtle
                -- 232 for Vashj'ir Seahorse (was named Abyssal Seahorse prior to Warlords of Draenor)
                -- 241 for Blue, Green, Red, and Yellow Qiraji Battle Tank (restricted to use inside Temple of Ahn'Qiraj)
                -- 242 for Swift Spectral Gryphon (hidden in the mount journal, used while dead in certain zones)
                -- 247 for Red Flying Cloud
                -- 248 for most flying mounts, including those that change capability based on riding skill
                -- 254 for Subdued Seahorse
                -- 269 for Azure and Crimson Water Strider
                if( mountType == 230 ) then
                    mountInfo.ground = true
                elseif( mountType == 231 ) then
                    mountInfo.ground = true
                elseif( mountType == 232 ) then
                    mountInfo.swimming = true
                    mountInfo.underwater = true
                elseif( mountType == 241 ) then
                    mountInfo.ground = true
                elseif( mountType == 242 ) then
                    mountInfo.ground = true
                    mountInfo.flying = true
                elseif( ( mountType == 247 ) or ( mountType == 248 ) ) then
                    mountInfo.ground = true
                    mountInfo.flying = true
                elseif( mountType == 254 ) then
                    mountInfo.underwater = true
                elseif( mountType == 269 ) then
                    mountInfo.ground = true
                    mountInfo.waterwalking = true
                end--if
            end--if
        end--if
    end--for
    return returnList;
end--LibPetAndMountDatabase:GetPlayersMounts()

function LibPetAndMountDatabase:GetPlayersCompanions( ctype )
    if( ctype == "BATTLEPET" ) then
        return self:GetPlayersBattleCompanions()
    elseif( ctype == "MOUNT" ) then
        return self:GetPlayersMounts()
    end--if
    return nil
end--LibPetAndMountDatabase:GetPlayersCompanions( ctype )

function LibPetAndMountDatabase:GetSummonedBattleCompanionIndex()
    local guid = C_PetJournal.GetSummonedPetGUID()

    return guid and tonumber( guid:sub( 3 ), 16 ) or nil
end--LibPetAndMountDatabase:GetSummonedBattleCompanionIndex()

function LibPetAndMountDatabase:GetSummonedMountIndex()
    for i = 1, C_MountJournal.GetNumMounts() do
        if( select( 4, C_MountJournal.GetMountInfo( i ) ) ) then
            return i
        end--if
    end--for
    return nil
end--LibPetAndMountDatabase:GetSummonedBattleCompanionIndex()

function LibPetAndMountDatabase:GetSummonedCompanionIndex( ctype )
    if( ctype == "BATTLEPET" ) then
        return self:GetSummonedBattleCompanionIndex()
    elseif( ctype == "MOUNT" ) then
        return self:GetSummonedMountIndex()
    end--if
    return nil
end--LibPetAndMountDatabase:GetSummonedCompanionIndex( ctype )

function LibPetAndMountDatabase:FindBattleCompanion( spellID )
    if( spellID and ( spellID < 0 ) ) then
        spellID = spellID * -1
    end--if
    if( spellID ) then
        local numPets, numOwned = C_PetJournal.GetNumPets()

        for i = 1, numOwned do
            local petID, _, isOwned = C_PetJournal.GetPetInfoByIndex( i )

            petID = tonumber( petID:sub( 3 ), 16 )
            if( isOwned and ( petID == spellID ) ) then
                return i;
            end--if
        end--for
    end--if
    return nil
end--LibPetAndMountDatabase:FindBattleCompanion( spellID )

function LibPetAndMountDatabase:FindMount( spellID )
    if( spellID and ( spellID < 0 ) ) then
        spellID = spellID * -1
    end--if
    if( spellID ) then
        for i = 1, C_MountJournal.GetNumMounts() do
            if( select( 2, C_MountJournal.GetMountInfo( i ) ) == spellID ) then
                return i
            end--if
        end--for
    end--if
    return nil
end--LibPetAndMountDatabase:FindMount( spellID )

function LibPetAndMountDatabase:FindCompanion( ctype, spellID )
    if( ctype == "BATTLEPET" ) then
        return self:FindBattleCompanion()
    elseif( ctype == "MOUNT" ) then
        return self:FindMount()
    end--if
    return nil
end--LibPetAndMountDatabase:FindCompanion( ctype, spellID )

function LibPetAndMountDatabase:MountInSet( item, set, normalize )
    local attributes = self:IdInSet( item, set, normalize )

    return attributes and ( type( attributes ) == "string" ) and self.GetMountAttributes( attributes ) or false
end--LibPetAndMountDatabase:MountInSet( item, set, normalize )
